/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "atemu.h"
#include "breakpoints.h"
#include "chip_data.h"
#include "chip_data_struct.h"

/* are the auto breakpoints set? */
int auto_breaks_active;

/* functions for using breakpoints */
void add_asm_breakpoint(int chipid, long addr){
    asm_break_tuple *to_add;
    char hex[32];
    GtkTreeIter iter;
    GtkTextIter start, end;
    long int line;
    
    gtk_list_store_append( chip_array[chipid].breakp, &iter );
    
    to_add = malloc(sizeof(asm_break_tuple));
    to_add->addr = addr;
    snprintf( hex, 32, "0x%08lx", addr );
    hex[31] = '\0';
 
    gtk_list_store_set( GTK_LIST_STORE(chip_array[chipid].breakp), &iter,
        0, "ASM", 1, hex, 2, addr, -1 );
    table_store(chip_array[chipid].asm_break, to_add);
    atemu_add_break(chipid, addr);

    line = address_to_text_line( chipid, addr );
    gtk_text_buffer_get_iter_at_line( chip_array[chipid].code,
            &start, line );
    gtk_text_buffer_get_iter_at_line( chip_array[chipid].code,
            &end, line + 1);
    gtk_text_buffer_apply_tag_by_name(chip_array[chipid].code,
            "breakpoint", &start, &end);
}

void add_source_breakpoint(int chipid, long addr, char *cfile){
    source_break_tuple *to_add;
    long int line;
    int sizeoffilename;
    char *filename;
    GtkTreeIter iter;
    GtkTextIter start, end;
    GtkTextBuffer *buff;

    // enough room for string, 6 digits, : and terminating \0
    sizeoffilename = strlen(cfile) + 6 + 2;
    filename = malloc(sizeoffilename);
    
    buff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(chip_array[curr_chip].textview_source));
    
    gtk_list_store_append( chip_array[chipid].breakp, &iter );
    
    to_add = malloc(sizeof(source_break_tuple));
    to_add->addr = addr;
    
    line = address_to_c_line( chipid, addr, &cfile );
    
    snprintf( filename, sizeoffilename,  "%s:%ld", cfile, line );
    filename[sizeoffilename-1] = '\0';
    
    gtk_list_store_set( GTK_LIST_STORE(chip_array[chipid].breakp), &iter,
        0, "Source", 1, filename, 2, addr, -1 );
    
    table_store(chip_array[chipid].source_break, to_add);
    atemu_add_break(chipid, addr);
    /* should also highlight a breakpointed line */
    gtk_text_buffer_get_iter_at_line(buff, &start, line - 1);
    gtk_text_buffer_get_iter_at_line(buff, &end, line);
    gtk_text_buffer_apply_tag_by_name(buff, "breakpoint", &start, &end);
}

void add_auto_breakpoint(int chipid, long addr){
    auto_break_tuple *to_add;

    to_add = malloc(sizeof(auto_break_tuple));
    to_add->addr = addr;
    
    table_store(chip_array[chipid].auto_break, to_add);
}

void remove_asm_breakpoint(int chipid, long addr){
    asm_break_tuple query;
    GtkTreeIter iter;
    GtkTextIter start, end;
    char *type;
    int line;
    long addrinnode;
    gboolean iterexists;
    
    query.addr = addr;
    iterexists = gtk_tree_model_get_iter_first( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter );

    line = address_to_text_line( chipid, addr );
    
    while( iterexists )
    {
        gtk_tree_model_get( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter, 0, &type,
            2, &addrinnode, -1 );
        if( !strncmp( type, "ASM", 3 ) )
        {
            if( addrinnode == addr )
            {
                gtk_list_store_remove( GTK_LIST_STORE(chip_array[chipid].breakp),
                    &iter );
                break;
            }
        }
        iterexists = gtk_tree_model_iter_next( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter );
    }
        
    table_delete(chip_array[chipid].asm_break, &query, free);
    atemu_clear_break(chipid, addr);

    /* should also un-highlight a (now) normal line */
    gtk_text_buffer_get_iter_at_line(chip_array[curr_chip].code, &start, line);
    gtk_text_buffer_get_iter_at_line(chip_array[curr_chip].code, &end, line + 1);
    gtk_text_buffer_remove_tag_by_name(chip_array[curr_chip].code, "breakpoint", &start, &end);
}

void remove_source_breakpoint(int chipid, long addr){
    source_break_tuple query;
    GtkTreeIter iter;
    GtkTextIter start, end;
    GtkTextBuffer *buff;
    long int line;
    char *type;
    char *filename;
    long addrinnode;
    gboolean iterexists;

    query.addr = addr;
    iterexists = gtk_tree_model_get_iter_first( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter );
    
    while( iterexists )
    {
        gtk_tree_model_get( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter, 0, &type,
            1, &filename, 2, &addrinnode, -1 );
        if( !strncmp( type, "Source", 3 ) )
        {
            if( addrinnode == addr )
            {
                /*take the line number off the filename*/
                *(strrchr( filename, ':' )) = '\0';
                buff = cfile_to_textbuffer( chipid, filename );
                line = address_to_c_line( chipid, addr, &filename );
                free(filename);
                gtk_list_store_remove( GTK_LIST_STORE(chip_array[chipid].breakp),
                    &iter );
                break;
            }
        }
        iterexists = gtk_tree_model_iter_next( GTK_TREE_MODEL(chip_array[chipid].breakp), &iter );
    }
    table_delete(chip_array[chipid].source_break, &query, free);
    atemu_clear_break(chipid, addr);

    /* should also un-highlight a (now) normal line */
    gtk_text_buffer_get_iter_at_line(buff, &start, line - 1);
    gtk_text_buffer_get_iter_at_line(buff, &end, line);
    gtk_text_buffer_remove_tag_by_name(buff, "breakpoint", &start, &end);
}

void remove_auto_breakpoint(int chipid, long addr){
    auto_break_tuple query;

    query.addr = addr;
    
    table_delete(chip_array[chipid].auto_break, &query, free);
}

/* return true if there is a breakpoint at the address */
int is_asm_breakpoint(int chipid, long addr){
    asm_break_tuple query;

    query.addr = addr;
    
    return (table_find(chip_array[chipid].asm_break, &query) != NULL);
}

int is_source_breakpoint(int chipid, long addr){
    source_break_tuple query;

    query.addr = addr;
    
    return (table_find(chip_array[chipid].source_break, &query) != NULL);
}

int is_auto_breakpoint(int chipid, long addr){
    auto_break_tuple query;

    query.addr = addr;
    
    return (table_find(chip_array[chipid].auto_break, &query) != NULL);
}


/* these functions make the list of automatic breakpoints.
 * later they will be activated and deactivated for C stepping */

void generate_auto_breaks_foreach(void *a, void *data){
    struct tuple_file_addr *at = (struct tuple_file_addr *)a;
    int *chipid = (int *)data;

    if(at->fname != NULL && cfile_to_textbuffer(*chipid, at->fname) != NULL){
        add_auto_breakpoint(*chipid, at->addr_begin);
    }
}

void generate_auto_breakpoints(int chipid){
    table_foreach(chip_array[chipid].addr_to_c_line, generate_auto_breaks_foreach, &chipid);
}

void activate_auto_breakpoints_foreach(void *a, void *data){
    auto_break_tuple *at = (auto_break_tuple *)a;
    int *chipid = (int *)data;

    atemu_add_break(*chipid, at->addr);

    //XXX something needs to be done to prevent pipe-full-deadlock
}

void activate_auto_breakpoints(int chipid){
    if(!auto_breaks_active){
        table_foreach(chip_array[chipid].auto_break, activate_auto_breakpoints_foreach, &chipid);
        auto_breaks_active = 1;
    }
}

void deactivate_auto_breakpoints_foreach(void *a, void *data){
    auto_break_tuple *at = (auto_break_tuple *)a;
    int *chipid = (int *)data;

    atemu_clear_break(*chipid, at->addr);
}

void deactivate_auto_breakpoints(int chipid){
    if(auto_breaks_active){
        table_foreach(chip_array[chipid].auto_break, deactivate_auto_breakpoints_foreach, &chipid);
        auto_breaks_active = 0;
    }
}

/* breakpoint comparison functions */
/* I probably don't need all three */

int asm_break_comp(const void *a, const void *b){
    asm_break_tuple *at, *bt;
    at = (asm_break_tuple*)a;
    bt = (asm_break_tuple*)b;

    return at->addr - bt->addr;
}

int source_break_comp(const void *a, const void *b){
    source_break_tuple *at, *bt;
    at = (source_break_tuple*)a;
    bt = (source_break_tuple*)b;

    return at->addr - bt->addr;
}

int auto_break_comp(const void *a, const void *b){
    auto_break_tuple *at, *bt;
    at = (auto_break_tuple*)a;
    bt = (auto_break_tuple*)b;

    return at->addr - bt->addr;
}

