/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>

#include "func_data.h"

struct func_data *create_func_data( char * name, int beg, int end )
{
    struct func_data *ret;

    ret = (struct func_data *) malloc( sizeof(struct func_data) );
    if ( ret == NULL )
    {
        return NULL;
    }

    /* Empty list */
    ret->locals = NULL;
    ret->c_lines = NULL;

    ret->name = strdup( name );
    ret->addr_begin = beg;
    ret->addr_end = end;

    return ret;
}

void add_c_line_to_func( struct func_data * fd, struct c_line * cl )
{
    if (( fd == NULL ) || ( cl == NULL ))
    {
        printf( "Passed NULL arg to add_c_line_to_func.\n" );
        return;
    }

    fd->c_lines = g_slist_append( fd->c_lines, (gpointer) cl );
}

void add_local_to_func( struct func_data * fd, struct local * loc )
{
    if (( fd == NULL ) || ( loc == NULL ))
    {
        printf( "Passed NULL arg to add_local_to_func.\n" );
        return;
    }

    fd->locals = g_slist_append( fd->locals, (gpointer) loc );
}

int func_data_compare( const void * a, const void * b )
{
    struct func_data *afd = (struct func_data *) a;
    struct func_data *bfd = (struct func_data *) b;

    /* see chip_data:addr_compare_range */

    if ( bfd->addr_begin > afd->addr_begin )
        return -1;

    if ( bfd->addr_end <= afd->addr_begin )
        return 1;

    return 0;    
}

void free_func_data( void * a )
{
    struct func_data *fd = (struct func_data *) a;
    GSList *curr;

    if ( fd == NULL )
        return;

    for( curr = fd->locals; curr != NULL; curr = curr->next )
        free_local( curr->data );

    g_slist_free( fd->locals );

    for( curr = fd->c_lines; curr != NULL; curr = curr->next )
        free_c_line( curr->data );

    g_slist_free( fd->c_lines );

    free( fd->name );
    free( fd );
}

struct local * create_local( char * name, int is_param, int type,
    int location, struct type_data *td, int id )
{
    struct local * ret;

    ret = (struct local *) malloc( sizeof(struct local) );
    if ( ret == NULL )
    {
        return NULL;
    }

    ret->name = strdup( name );
    ret->is_param = is_param;
    ret->type = type;
    ret->location = location;
    ret->td = td;
    ret->id = id;

    if ( td == NULL )
    {
        ret->value = calloc( 4, sizeof( uint8_t ) );
    } else {
        ret->value = calloc( td->byte_length, sizeof( uint8_t ) );
    }

    ret->type_style = 0;

    return ret;
}

void free_local( struct local * loc )
{
    if ( loc )
        free( loc->name );

    free( loc );
}

struct c_line * create_c_line( long int addr, long int line )
{
    struct c_line *ret;

    ret = (struct c_line *) malloc( sizeof(struct c_line) );
    if ( ret == NULL )
    {
        return NULL;
    }
 
    ret->addr = addr;
    ret->line = line;
 
    return ret; 
}

void free_c_line( struct c_line * cl )
{
    free( cl );
}
